/*	This file is part of the auxiliaries library.
	Written by Dick Grune, Vrije Universiteit, Amsterdam.
	$Id: ls.c,v 1.20 2019-10-25 09:43:41 dick Exp $
*/

#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<time.h>
#include	<string.h>

#ifdef	UNIX
#include	<pwd.h>
#include	<grp.h>
#endif	/* MSDOS */

#include	"ls.h"

/* Library module source prelude */
#undef	_LS_CODE_
#ifndef	lint
#define	_LS_CODE_
#endif
#ifdef	LIB
#define	_LS_CODE_
#endif

#ifdef	_LS_CODE_

/* Library module source code */

							/* AUXILIARIES */

static int
charinset(const char *set, char ch) {
	if (!set) return 0;
	while (*set && *set != ch) {
		set++;
	}
	return *set == ch;
}

static const char *
base_pos(const char *name) {
	const char *res = name;		/* default */

	while (*name) {
		if (*name == '/' || *name == '\\') {
			res = name+1;
		}
		name++;
	}
	return res;
}

static void
print_basename(const char *name) {
	printf(base_pos(name));
}

static void
prperms(unsigned int perm) {
	unsigned int prm = perm;
	char pp[9];
	char *p = pp;
	int i;

	for (i = 0; i < 3; i++) {
#ifndef	MSDOS
		*p++ = prm & S_IREAD ? 'r' : '-';
		*p++ = prm & S_IWRITE ? 'w' : '-';
		*p++ = prm & S_IEXEC ? 'x' : '-';
#else
		*p++ = prm & S_IRUSR ? 'r' : '-';
		*p++ = prm & S_IWUSR ? 'w' : '-';
		*p++ = 'x';
#endif
		prm <<= 3;
	}
#ifdef	UNIX
	if (perm & S_ISUID) {
		pp[2] = 's';
	}
	if (perm & S_ISGID) {
		pp[5] = 's';
	}
	if (perm & S_ISVTX) {
		pp[8] = 't';
	}
#endif
	printf("%.9s", pp);
}

static void
print_dirname(const char *name) {
	const char *bp = base_pos(name);
	while (*name && name != bp) {
		printf("%c", *name++);
	}
}

							/* THE ENTRIES */
int
ls(const char *name, const struct stat *buf, const char *flags) {
	if (!name) return 0;

	unsigned int mode = (buf ? buf->st_mode : 0);
	unsigned int type = mode & S_IFMT;
	if (charinset(flags, 'd') && type != S_IFDIR) return 0;

	if (charinset(flags, 'r')) {
		print_basename(name);
		printf("\t");
	}

	if (charinset(flags, 'l')) {
		if (charinset(flags, 'i')) {
			printf("%5u ", (buf ? (unsigned int)buf->st_ino : 0));
		}

		printf("%c",
			type == S_IFCHR	? 'c' :
			type == S_IFDIR	? 'd' :
			type == S_IFREG	? '-' :
#ifdef	S_IFBLK
			type == S_IFBLK	? 'b' :
#endif
#ifdef	S_IFMPC
			type == S_IFMPC ? 'm' :
#endif
#ifdef	S_IFMPB
			type == S_IFMPB ? 'M' :
#endif
#ifdef	S_IFIFO
			type == S_IFIFO ? 'f' :
#endif
#ifdef	S_IFLNK
			type == S_IFLNK ? 'l' :
#endif
#ifdef	S_IFSOCK
			type == S_IFSOCK? 's' :
#endif
						'?');

		prperms(mode);

		printf(" %3d", (buf ? (int)buf->st_nlink : 0));

#ifdef	UNIX
		if (buf) {
			struct passwd *p;
			setpwent();
			p = getpwuid(buf->st_uid);
			if (p != 0) {
				printf(" %8.8s", p->pw_name);
			}
			else {
				printf(" %8d", (int)buf->st_uid);
			}
		} else {
			printf(" --------");
		}

		if (buf) {
			struct group *p;
			setgrent();
			p = getgrgid(buf->st_gid);
			if (p != (struct group *)0) {
				printf(" %8.8s", p->gr_name);
			}
			else {
				printf(" %8d", (int)buf->st_gid);
			}
		} else {
			printf(" --------");
		}
#endif

		printf(" %10lu", (buf ? buf->st_size : 0));

		{
			char *tm = ctime(buf ? &buf->st_mtime : 0);
			if (tm) {
				/* remove trailing \n */
				tm[24] = '\0';
			} else {
				/* ctime() failed */
				tm = "????????????????????????";
			}
			printf(" %s ", &tm[4]);
		}
	}
	else if (charinset(flags, 's')) {
		printf(" %10lu ", (buf ? buf->st_size : 0));
	} else {
	}

	if (charinset(flags, 'r')) {
		print_dirname(name);
	} else {
		printf("%s", name);
		if (type == S_IFDIR) {
			if (name[strlen(name)-1] != '/') {
				printf("/");
			}
		}
	}

	return 1;
}

/* End library module source code */
#endif	/* _LS_CODE_ */

#ifdef	lint
static void
satisfy_lint(void *x) {
	/* not needed because ls.h defines only one entry */
	satisfy_lint(x);
}
#endif	/* lint */
